//$(
//=======================================================================
//
//      File: $RCSfile: watchdogRT.c,v $
//      Version: $Revision: $  Dated: $Date: $
//
//=======================================================================
//
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version
//	2 of the License, or (at your option) any later version.
//
//	The author, Walter S. Heath, will not admit liability nor
//	provide warranty for any of this software. This material is
//	provided "AS-IS" and at no charge.
//
//	(c) Copyright 1999    Walter S. Heath,  walterh@ll.mit.edu
//
//	Release 1.0
//
//  FILE: watchdogRT.c
//
//  DESCRIPTION:
//
//	A program to initialize and manage a PC watchdog timer. The
//	program uses the alarm interrupt feature of the PC's battery-
//	backed CMOS Real-Time Clock (RTC). It uses the Linux /dev/rtc
//	device to access the RTC. When running, it pushes the alarm
//	timeout time ahead periodically such that the timeout time
//	is never reached. If the program should terminate abnormally
//	or be blocked from operation by some other program (higher
//	priority), the timeout time will be reached and the RTC chip
//	will generate an IRQ8 interrupt. On some embedded PC boards,
//	including the AMPRO Little Board /P5e PC, this interrupt may be
//	jumpered (jumper W24) to cause a board RESET, which, in turn
//	will cause a re-boot of Linux.
//
//	The program may be terminated gracefully (without alarm timeout)
//	by entering Ctrl-C. To test the alarm timeout feature, it may
//	be terminated with a SIGQUIT signal (e.g., kill -3 PID). That
//	signal will leave the timeout timer running and an IRQ8 interrupt
//	will occur.
//
//	File rt_watchdog.c contains a thread that does most of the
//	work. It could be added to a multi-threaded (even real-time)
//	program as the lowest priority thread to manage an auto-reboot
//	of the program. Alternatively, this program may be started as-is
//	and allowed to run concurrently with other application
//	processes (probably at a lower priority). It will then manage
//	the RTC timeout as long as it isn't blocked from running for
//	too long (about five seconds, as currently configured.)
//
//	This program also demonstrates a general-purpose technique for
//	initializing and starting a multi-threaded program such that
//	all threads are initialized before any thread begins its
//	processing loop. This particular program only starts one
//	additional thread. But the technique is easily generalized for
//	a multi-threaded program. It has been used to start Linux
//	real-time threads. An excellent reference for POSIX thread
//	programming is:
//
//	"Programming with POSIX(R) Threads", by David R. Butenhof
//	Wddison-Wesley Publishing Company, (C) 1997, ISBN 0-201-63392-2.
//
//	To run this program at real-time scheduling priority, change
//	the set_thread_attr() call in this file to:
//
//		set_thread_attr( &thread_attr, SCHED_RR, 3 );
//
//	The thread will then be scheduled using real-time Round Robin
//	scheduling at priority 3. You must run real-time programs at
//	root level.
//
//	This program has been successfully compiled under redhat(R) Linux
//	5.2 and 6.0 (www.redhat.com). It may be compiled as a C (gcc)
//	or C++ (g++) program.  Its ability to cause a timeout RESET
//	has been demonstrated on an AMPRO Little Board /P5e processor
//	(www.ampro.com). It may also be compiled and run on a standard
//	desktop PC. It will cause an IRQ8 interrupt when it times out,
//	but most desktop PCs (and Linux) will ignore this interrupt. It
//	may be possible to wire the IRQ8 interrupt to the RESET line on a
//	desktop motherboard (with suitable pulse conditioning -- I would
//	appreciate hearing if someone has experience with this). To see
//	what's happening, compile the program with -DDEBUG active. Make
//	sure your kernel contains the /dev/rtc device.
//
// LINUX "BUG" WORK-AROUND:
//
//	This is a problem found in Linux 2.2: During the boot sequence,
//	the /etc/rc.d/rc.sysinit file runs: "hwclock --hctosys" to
//	initialize the system clock (sys) from the hardware clock (hc).
//	It causes an IRQ8 interrupt! If IRQ8 is jumpered to cause
//	a RESET, the system is reset and reboots -- not what we want!
//
//	The file must be changed to run: "hwclock --hctosys --directisa".
//	It then does not use the /dev/rtc driver. Instead, it does
//	direct I/O (i.e., inb(), outb() to ports 0x70, 0x71) to read
//	the RTC time/date. This method does not cause an IRQ8 interrupt.
//
//	See: "man hwclock" for details. It indicates that the driver
//	for /dev/rtc is relatively new. Whether the IRQ8 interrupt is
//	a bug or a necessity is a question for someone else to answer.
//	Until then, you need to change the rc.sysinit file as indicated
//	to use this watchdog program.
//
//
//$)
//

#include <watchdog.h>
#include <signal.h>

pthread_t thread_id;

startup_t startup = { PTHREAD_MUTEX_INITIALIZER,
				PTHREAD_COND_INITIALIZER, 0 };

/* All process globals defined here */
int thread_ready = 0;


/* Local prototypes */
static void set_thread_attr( pthread_attr_t *thread_attr_p,
					int policy, int priority );
static void WatchdogCleanup( int i);
static void AbortCleanup( int i );



int main( int argc, char *argv[] )
{
    pthread_attr_t thread_attr;
    int status;

    signal( SIGINT, WatchdogCleanup ); /* catch Ctrl-C signal */
    signal( SIGQUIT, AbortCleanup );   /* catch a SIGQUIT(=3) signal */

    /* Set thread attributes (could be SCHED_RR for real-time */
    set_thread_attr( &thread_attr, SCHED_OTHER, 1 );


    /* Create thread -- don't detach it */
    status = pthread_create( &thread_id, &thread_attr,
                                rtWatchdogThread, NULL );
    if( status != 0 )
        err_abort( status, "Can\'t create thread 1" );


    /* Wait for thread to be ready */
    while( 1 ) {

        if( thread_ready )
            break;

        sched_yield( );
    }

    /* Broadcast "go" condition */
    status = pthread_mutex_lock( &startup.mutex );
    if( status != 0 )
        err_abort( status, "Can\'t lock startup.mutex" );

    startup.go = 1;

    status = pthread_cond_broadcast( &startup.cond );
    if( status != 0 )
        err_abort( status, "Can\'t broadcast startup.cond" );

    status = pthread_mutex_unlock( &startup.mutex );
    if( status != 0 )
        err_abort( status, "Can\'t unlock startup.mutex" );

    /* Just sleep for awhile */
    for( ; ; )
	sleep( 100 );

}



//
//   set_thread_attr()
//
// Routine to set a thread's attributes ( policy and priority ).
// 'priority' has range: 1 (lowest) to N (highest). N is 99 for SCHED_RR
// and SCHED_FIFO policy, 1 for SCHED_OTHER (non-real-time).
//
//

static void set_thread_attr( pthread_attr_t *thread_attr_p,
					 int policy, int priority )
{
    struct sched_param thread_param;
    int status, min_priority, max_priority;
    int thread_policy;

    /* Initialize attribute structure */
    status = pthread_attr_init( thread_attr_p );
    if( status != 0 )
	err_abort( status, "Can\'t init attr" );

#undef DEBUG

    /* Make sure priority scheduling is supported! */
#if defined (_POSIX_THREAD_PRIORITY_SCHEDULING ) && !defined (sun)

 #ifdef DEBUG
    status = pthread_attr_getschedpolicy( thread_attr_p, &thread_policy );
    if( status != 0 )
	err_abort( status, "Can\'t get policy" );

    status = pthread_attr_getschedparam( thread_attr_p, &thread_param );
    if( status != 0 )
	err_abort( status, "Can\' get sched param" );

    printf( "Default policy: %s, priority: %d\n",
	( thread_policy == SCHED_FIFO ? "FIFO"
	: ( thread_policy == SCHED_RR ? "RR"
	  : ( thread_policy == SCHED_OTHER ? "OTHER"
	    : "unknown" ))), thread_param.sched_priority );
 #endif
    /* Set scheduling policy */
    status = pthread_attr_setschedpolicy( thread_attr_p, policy );
    if( status != 0 )
	err_abort( status, "Can\'t set policy" );

    /* Set scheduling priority */
    if( priority < 1 ) {
	fprintf( stderr, "Priority value: %d can\'t be below 1\n" );
	abort();
    }
    min_priority = sched_get_priority_min( policy );
    if( min_priority == -1 )
	errno_abort( "Can\'t get min priority" );

    max_priority = sched_get_priority_max( policy );
    if( max_priority == -1 )
	errno_abort( "Can\'t get max priority" );

    thread_param.sched_priority = min_priority + (priority - 1);
    if( thread_param.sched_priority > max_priority ) {
	fprintf( stderr, "Selected priority is higher than max allowed\n" );
	abort();
    }

  #ifdef DEBUG
    printf( "Priority range is: %d to %d: using %d\n", 
	min_priority, max_priority, thread_param.sched_priority );
  #endif

    /* Set scheduling parameters */
    status = pthread_attr_setschedparam( thread_attr_p, &thread_param );
    if( status != 0 )
	err_abort( status, "Can\'t set params" );

  #ifdef DEBUG
    printf( "Setting explicit scheduling at priority: %d\n",
					thread_param.sched_priority );
  #endif

    /* Tell it to use explicit (not enherited) scheduling */
    status = pthread_attr_setinheritsched( thread_attr_p,
					PTHREAD_EXPLICIT_SCHED );
    if( status != 0 )
	err_abort( status, "Can\'t set explicit scheduling" );

 #ifdef DEBUG
    status = pthread_attr_getschedpolicy( thread_attr_p, &thread_policy );
    if( status != 0 )
        err_abort( status, "Can\'t get policy" );

    status = pthread_attr_getschedparam( thread_attr_p, &thread_param );
    if( status != 0 )
        err_abort( status, "Can\' get sched param" );

    printf( "Changed policy: %s, priority: %d\n\n",
        ( thread_policy == SCHED_FIFO ? "FIFO"
        : ( thread_policy == SCHED_RR ? "RR"
          : ( thread_policy == SCHED_OTHER ? "OTHER"
            : "unknown" ))), thread_param.sched_priority );
 #endif

#else
//    printf("Priority scheduling not supported\n");
#endif
}


//
//  wait_for_go()
//
//  This routine is called by application threads to synchronize
//  startup.
//
void wait_for_go( int thread_number )
{
    struct timespec timeout;
    int status;

    timeout.tv_sec = time( NULL ) + 4;
    timeout.tv_nsec = 0;

    status = pthread_mutex_lock( &startup.mutex );
    if( status != 0 )
	err_abort( status, "Can\'t lock startup.mutex" );

    while( startup.go == 0 ) {
	status = pthread_cond_timedwait( &startup.cond, &startup.mutex,
							&timeout );
	if( status == ETIMEDOUT ) {
	    fprintf( stderr, "Startup timeout on thread %d\n",
						thread_number );
	    abort();
	}
	else if( status != 0 )
	    err_abort( status, "timewait error" );
    }

    status = pthread_mutex_unlock( &startup.mutex );
    if( status != 0 )
	err_abort( status, "Can\'t unlock startup.mutex" );
}


//
// SIGINT signal handler.
//
static void WatchdogCleanup( int i )
{
    int status;

    /* Cancel thread rtWatchdogThread (THREAD_1) */
    if( pthread_self() == thread_id ) {

	status = pthread_cancel( thread_id );
	if( status != 0 )
	    err_abort( status, "pthread_cancel() error" );


	status = pthread_join( thread_id, NULL );
	if( status != 0 )
	    err_abort( status, "pthread_join() error" );
    }

    exit( 0 );
}

//
//  SIGQUIT signal handler.
//
//  Used to demonstrate the RESET operation. Terminating
//  the program by entering:
//
//	kill -3 <process PID>
//
//  will leave the alarm timeout active. The PC will then
//  generate an IRQ8 interrupt when current time matches
//  the alarm timeout time.
//
static void AbortCleanup( int i )
{
    _exit( 1 );
}

